unit FmxCore;
/////////////////////////////////////////////////////////////////////////////
//
// Fidonet Mail Express: FMX Classes that provides object APIs
//                       for FMA.
//
// FmxCore.Pas
//
// Copyright (c) 2000-2001 by Akzhan Abdulin
//                                         http://www.akzhan.midi.ru/fma/
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version. See GNU_GPL.TXT for details.
//
// Created: 18/Feb/2000
// Updated: 14/Oct/2002
//
// History: 18/Feb/2000 akzhan  - Initially created for Delphi.
//          14/Feb/2001 akzhan  - Alpha 1.
//          15/Feb/2001 akzhan  - Beta (required functionality implemented)
//                                TFmxMessageObject.AddStringProperty fixed.
//                                TFmxMsgPropAccessor GetStrings and GetString API added.
//                                TFmxMessageObject.AddStringsProperty added.
//                                TFmxCustomAreaMessage Load, Save, SaveLen API added.
//                                Property navigation implemented.
//                                TFmxMessageObject.Properties implemented.
//          16/Feb/2001 akzhan  - Beta 2.
//                                TFmxMsgPropAccessor.Delete method implemented.
//                                TFmxMsgPropAccessor DataPointer, PropertyData,
//                                Id, Position and Tag properties added.
//          14/Oct/2002 Plyako A- Beta 3.
//                                TFmxPropAccessor.FindProperty method added.
//                                TFmxPropAccessor.Attributes is writable now.
//                                TFmxCustomAreaMessages.Read method implemented.
//                                TFmxCustomMessageArea.EnumMessages is a function now.
//                                TFmxCustomMessageArea.DoClose is a function now.
//                                Errors ignoring routines added.
//                                Some bugs fixed.
//
/////////////////////////////////////////////////////////////////////////////

{$I FmxDef.inc}

interface

uses
  SysUtils, Windows, Classes,
  FmaDef, FmaApi, FmxApi;

type
  EFmxError = class(Exception)
  private
    FErrorCode: TFmaError;
  public
    constructor Create;
    property ErrorCode: TFmaError read FErrorCode;
  end;

type
  TFmxMessageObject = class;
  TFmxMsgPropAccessor = class
  private
    FMsgObj: PMsgObject;
    FData: PFMP;
    function GetId: ULONG;
    procedure SetId(const Value: ULONG);
    function GetPosition: Byte;
    procedure SetPosition(const Value: Byte);
    function GetTag: UINT;
    procedure SetTag(const Value: UINT);
    function GetPropertyData: Pointer;
    function GetPSZCount: Integer;
    function GetByteCount: Integer;
    function GetByte(Index: Cardinal): BYTE;
    procedure SetByte(Index: Cardinal; Value: BYTE);
    function GetU32Count: Integer;
    function GetU32(Index: Cardinal): LONG;
    procedure SetU32(Index: Cardinal; Value: LONG);
    function GetDateTime: TDateTime;
    procedure SetDateTime(Value: TDateTime);
  public
    constructor CreateOnPFMP(AMsgObj: PMsgObject);
    constructor Create(AMsgObj: TFmxMessageObject);

    // Simplified API
    function Eol: Boolean;
    procedure First;
    procedure Last;
    procedure Next;
    procedure Prev;

    procedure Dump;

    procedure Delete;

    function IsASZ: Boolean;
    function IsPSZs: Boolean;
    function IsU32s: Boolean;
    function IsBytes: Boolean;
    function IsDateTime: Boolean;

    function IsEndOfMessage: Boolean;
    function IsSingleInstance: Boolean;
    function IsKludge: Boolean;

    function GetASZ: PAnsiChar;
    function GetPSZ(Index: Cardinal): PAnsiChar;
    procedure SetPSZ(Index: Cardinal; Value: AnsiString);

    // Simplified API
    function GetString: AnsiString;
    procedure GetStrings(List: TStrings);

    function FindProperty(const PropID: ULONG): Boolean;
    function FindKludge(Kludge: string): Boolean;

    property DataPointer: PFMP read FData;
    property PropertyData: Pointer read GetPropertyData;
    property Id: ULONG read GetId write SetId;
    property Position: Byte read GetPosition write SetPosition;
    property Tag: UINT read GetTag write SetTag;

    property PSZCount: Integer read GetPSZCount;
    property U32Count: Integer read GetU32Count;
    property AsU32S[Index: Cardinal]: LONG read GetU32 write SetU32;
    property ByteCount: Integer read GetByteCount;
    property AsBytes[Index: Cardinal]: BYTE read GetByte write SetByte;
    property AsDateTime: TDateTime read GetDateTime write SetDateTime;
  end;

  TFmxMessageObject = class
  private
    FData: PMsgObject;
    FAccessor: TFmxMsgPropAccessor;
    function GetPropAccessor: TFmxMsgPropAccessor;
    function GetNativeType: TFmxMsgType;
    function GetAttributes: TFmxMsgAttributes;
    procedure SetAttributes(const Value: TFmxMsgAttributes);
  public
    constructor CreateOn(AData: PMsgObject);
    constructor CreateNew(ExtraSize: UINT);
    constructor CreateClone(AnObject: TFmxMessageObject);
    destructor Destroy; override;

    procedure Dump;

    function GetFirstProperty: PFMP;
    function GetProperty(Id: ULONG; Index: UINT): PFMP;
{$IFDEF FMA_D5}
    function RemoveProperty(Id: ULONG; Index: UINT = 0): Boolean;
{$ELSE}
    function RemoveProperty(Id: ULONG; Index: UINT): Boolean;
{$ENDIF}
    procedure RemoveProperties(Id: ULONG);
    function AddProperty(Id: ULONG; Data: Pointer; Size: ULONG): PFMP;


    function AddASZProperty(Id: ULONG; Value: PAnsiChar): PFMP;
    function AddU32Property(Id: ULONG; Values: array of LONG): PFMP;
    function AddDTMProperty(Id: ULONG; Value: TDateTime): PFMP;
    function AddBINProperty(Id: ULONG; Values: array of BYTE): PFMP;
    function AddPSZProperty(Id: ULONG; Values: array of PAnsiChar): PFMP;
    function AddStringProperty(Id: ULONG; Value: AnsiString): PFMP;
{$IFDEF FMA_D5}
    function AddStringArrayProperty(Id: ULONG; Values: array of AnsiString): PFMP;
    function AddStringsProperty(Id: ULONG; List: TStrings): PFMP;
{$ENDIF}


    property DataPointer: PMsgObject read FData;
    property NativeType: TFmxMsgType read GetNativeType;
    property Attributes: TFmxMsgAttributes read GetAttributes write SetAttributes;
    property Properties: TFmxMsgPropAccessor read GetPropAccessor;
  end;

type
  TFmxCustomAreaMessage = class
  private
    FHandle: TMsgHandle;
    function GetControlInfoLen: ULONG;
    function GetTextLen: ULONG;
  protected
    procedure SetHandle(Value: TMsgHandle);

    function DoClose: Boolean; virtual; abstract;
    function DoGetControlInfoLen(var Len: ULONG): Boolean; virtual; abstract;
    function DoGetTextLen(var Len: ULONG): Boolean; virtual; abstract;

    function DoLoad(var ppfmm: PMsgObject; fl: ULONG): Boolean; virtual; abstract;
    function DoSave(pfmm: PMsgObject; fl: ULONG; cchTextTotal: ULONG): Boolean; virtual; abstract;

    function DoRead(pmsgh: Pointer; var ppchCtrl: PAnsiChar; pcchCtrl: PULONG; var ppchText: PAnsiChar; pcchText: PULONG; poffText: PULONG): Boolean; virtual; abstract;
    function DoWrite(pmsgh: Pointer; pchCtrl: PAnsiChar; cchCtrl: ULONG; pchText: PAnsiChar; cchText: ULONG; poffText: PULONG; cchTextTotal: ULONG): Boolean; virtual; abstract;
  public
    constructor CreateOn(MsgHandle: TMsgHandle);
    constructor Create(AreaHandle: TAreaHandle; mode: UINT; msgn: TMsgNumber; fl: ULONG); virtual; abstract;
    destructor Destroy; override;

    procedure Close;

    procedure Read(lines: TStrings);

    function Load(fl: ULONG): PMsgObject;
    procedure Save(pfmm: PMsgObject; fl: ULONG);
    procedure SaveLen(pfmm: PMsgObject; fl: ULONG; TextTotalLen: ULONG);

    property Handle: TMsgHandle read FHandle;
    property ControlInfoLength: ULONG read GetControlInfoLen;
    property TextLength: ULONG read GetTextLen;
  end;

type
  TFmxCustomMessageArea = class
  private
    FHandle: TAreaHandle;
    FPath: AnsiString;
    FOpenMode: TFmxAreaOpenMode;
    FNetAddr: TFmxNetAddr;
    FFlags: TFmxAreaOpenFlags;
    FUserId: ULONG;
    FUserName: AnsiString;

    function GetActive: Boolean;
    procedure SetActive(const Value: Boolean);

    procedure SetPath(const Value: String);
    procedure SetFlags(const Value: TFmxAreaOpenFlags);
    procedure SetNetAddr(const Value: TFmxNetAddr);
    procedure SetOpenMode(const Value: TFmxAreaOpenMode);

    function GetHighestReadMessageId: UMSG;
    function GetLastReadMessageId: UMSG;
    procedure SetHighestReadMessageId(const Value: UMSG);
    procedure SetLastReadMessageId(const Value: UMSG);
    function GetCurrentMessageNumber: TMsgNumber;
    procedure SetCurrentMessage(msgn: TMsgNumber);
    function GetMessageCount: LONG;

  protected
    procedure SetHandle(Value: TAreaHandle);

    procedure DoOpen(pszPath: PAnsiChar; fOpenMode: UINT; fl: ULONG; var pnetAddr: NETADDR); virtual; abstract;
    function DoClose: Boolean; virtual; abstract;

    function DoLock(LockArea: Boolean): Boolean; virtual; abstract;

    function DoGetLRInfo(var lri: LRINFO): Boolean; virtual; abstract;
    function DoSetLRInfo(const lri: LRINFO): Boolean; virtual; abstract;

    function DoValidateMessageNumber(MsgN: TMsgNumber; var MsgIndex: ULONG): Boolean; virtual; abstract;

    function DoCreateMsgMap(var paumsg: PUMSG; var pcumsg: ULONG): Boolean; virtual; abstract;

    function DoEnumMessages(CallProc: TEnumMsgsObjProc; Flags: TEnumMsgsFlags): Boolean; virtual; abstract;

    function DoGetCurrentMessage: TMsgNumber; virtual; abstract;
    function DoSetCurrentMessage(msgn: TMsgNumber): Boolean; virtual; abstract;
    function DoGetFirstMessage: TMsgNumber; virtual; abstract;
    function DoGetHighMessage: TMsgNumber; virtual; abstract;
    function DoGetMessageCount: LONG; virtual; abstract;
    function DoDeleteMessage(msgn: TMsgNumber): Boolean; virtual; abstract;

    function DoMsgIdToMsgNumber(Value: TMsgId; Mode: UINT): TMsgNumber; virtual; abstract;
    function DoMsgNumberToMsgId(Value: TMsgNumber): TMsgId; virtual; abstract;

    function DoGetMessage(msgn: TMsgNumber; Mode: UINT; Flags: ULONG): TFmxCustomAreaMessage; virtual; abstract;

  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure Open;
    procedure Close;

    procedure Lock;
    procedure Unlock;

    function ValidateMessageNumber(MsgN: TMsgNumber; var MsgIndex: ULONG): Boolean;
{$IFDEF FMA_D5}
    function GetMessageMap: TMsgIdArray;
{$ENDIF}
    function EnumMessages(CallProc: TEnumMsgsObjProc; Flags: TEnumMsgsFlags): Boolean;
    function GetFirstMessageNumber: TMsgNumber;
    function GetHighMessageNumber: TMsgNumber;
    procedure DeleteMessage(msgn: TMsgNumber);

    function GetMessage(msgn: TMsgNumber; Mode: TFmxMessageOpenMode; Flags: TFmxMessageOpenFlags): TFmxCustomAreaMessage;

    // Simplified API
    procedure First;
    procedure Prev;
    procedure Next;
    procedure Last;
    function Validate: ULONG;
    procedure Delete;
    function Get(Mode: TFmxMessageOpenMode; Flags: TFmxMessageOpenFlags): TFmxCustomAreaMessage;

    function MsgIdToMsgNumber(Value: TMsgId; Mode: TFmxMsgIdConvertionMode): TMsgNumber;
    function MsgNumberToMsgId(Value: TMsgNumber): TMsgId;

    property Handle: TAreaHandle read FHandle;
    property Active: Boolean read GetActive write SetActive default False;

    property Path: String read FPath write SetPath;
    property OpenMode: TFmxAreaOpenMode read FOpenMode write SetOpenMode default aomExisting;
    property NetAddr: TFmxNetAddr read FNetAddr write SetNetAddr;
    property OpenFlags: TFmxAreaOpenFlags read FFlags write SetFlags default [];

    property UserId: ULONG read FUserId write FUserId default 0;
    property UserName: AnsiString read FUserName write FUserName;
    property LastReadMessageId: UMSG read GetLastReadMessageId write SetLastReadMessageId stored False;

    property HighestReadMessageId: UMSG read GetHighestReadMessageId write SetHighestReadMessageId stored False;

    property MessageNumber: TMsgNumber read GetCurrentMessageNumber write SetCurrentMessage stored False;
    property MessageCount: LONG read GetMessageCount;
  end;

// Ignore error err
procedure FmxDeactivateError(err: TFmaError);
// Do not ignore error err
procedure FmxActivateError(err: TFmaError);
// Is error err ignored?
function  FmxErrorIsActive(err: TFmaError): Boolean;
// NB (Andrew D. Plyako)



procedure FmxErrorCheck;
procedure FmxErrorRaise(Err: TFmaError);
procedure FmxErrorCheckExcept(SkipErrors: array of const);

implementation
{$IFNDEF VIRTUALPASCAL}

{$BOOLEVAL OFF}

{$RANGECHECKS OFF}
{$ENDIF}

uses
  FmxDef, FmaSys, FmxRes
  {$IFDEF FMA_STATIC_LINKAGE}, FmaObj{$ENDIF};

type
  APAnsiChar = array[0..0] of PAnsiChar;
  ALONG = array[0..0] of LONG;

var
  IgnoreErrors: set of byte;

procedure FmxDeactivateError(err: TFmaError);
begin
  if ord(err)<=High(Byte) then
  Include(IgnoreErrors, ord(err));
end;

procedure FmxActivateError(err: TFmaError);
begin
  if ord(err)<=High(Byte) then
  Exclude(IgnoreErrors, ord(err));
end;

function  FmxErrorIsActive(err: TFmaError): Boolean;
begin
  Result := (ord(err)>High(Byte)) or not (ord(Err) in IgnoreErrors);
end;

procedure FmxErrorCheck;
var LastError: TFmaError;
begin
  LastError := FmaGetLastError();
  if (LastError <> FMAERROR_NONE) and FmxErrorIsActive(LastError) then
    raise EFmxError.Create
  else FmaSetLastError(FMAERROR_NONE)
end;

procedure FmxErrorRaise(Err: TFmaError);
begin
  FmaSetLastError(Err);
  FmxErrorCheck;
end;

procedure FmxErrorCheckExcept(SkipErrors: array of const);
var
  I: Integer;
  Err: TFmaError;
begin
  Err := FmaGetLastError;
  if Err <> FMAERROR_NONE then
  begin
    for I := Low(SkipErrors) to High(SkipErrors) do
    begin
      if
        ((SkipErrors[I].VType = vtInteger) and (Err = TFmaError(SkipErrors[I].VInteger)))
{$IFDEF FMA_D5}
      or
        ((SkipErrors[I].VType = vtInt64) and (Err = TFmaError(SkipErrors[I].VInt64)))
{$ENDIF}
      then
      begin
        Exit;
      end;
    end;
    raise EFmxError.Create;
  end;
end;

{ EFmxError }

constructor EFmxError.Create;
var
  Err: TFmaError;
begin
  Err := FmaGetLastError;
  Inherited Create(FmxGetErrorDescription(Err));
  FErrorCode := Err;
end;

{ TFmxCustomMessageArea }

{
  NB (Andrew D. Plyako)
  -     FmaCloseArea
  - Properly check FmaCloseArea result
}
procedure TFmxCustomMessageArea.Close;
var AllOk: Boolean;
begin
  if Handle <> NoAreaHandle then
  begin
    AllOk := DoClose;
    FHandle := NoAreaHandle;
    if not AllOk then FmxErrorCheck;
  end;
end;
// NB (Andrew D. Plyako)

constructor TFmxCustomMessageArea.Create;
begin
  Inherited Create;
  FHandle := NoAreaHandle;
  FPath := '';
  FOpenMode := aomExisting;
  FNetAddr := NoNetAddr;
  FFlags := [];
  FUserId := 0;
  FUserName := '';
end;

procedure TFmxCustomMessageArea.Delete;
begin
  DeleteMessage(MessageNumber);
end;

procedure TFmxCustomMessageArea.DeleteMessage(msgn: TMsgNumber);
begin
  if not DoDeleteMessage(msgn) then
  begin
    FmxErrorCheck;
  end;
end;

destructor TFmxCustomMessageArea.Destroy;
begin
  Close;
  inherited Destroy;
end;


{
  NB (Andrew D. Plyako)
  -  ,      ""
  - method "EnumMessages" is a function now. If you cancel enumeration in
    CallProc then EnumMessages return true
}
function TFmxCustomMessageArea.EnumMessages(CallProc: TEnumMsgsObjProc;
  Flags: TEnumMsgsFlags): Boolean;
begin
  Result := not DoEnumMessages(CallProc, Flags);
  if Result then // Ether error occurred or Enumeration canceled by user
     FmxErrorCheck;
end;
// NB (Andrew D. Plyako)



procedure TFmxCustomMessageArea.First;
begin
  SetCurrentMessage(1);
end;

function TFmxCustomMessageArea.Get(Mode: TFmxMessageOpenMode;
  Flags: TFmxMessageOpenFlags): TFmxCustomAreaMessage;
begin
  Result := GetMessage(TMsgNumber_CUR, Mode, Flags);
end;

function TFmxCustomMessageArea.GetActive: Boolean;
begin
  Result := (Handle <> NoAreaHandle);
end;

function TFmxCustomMessageArea.GetCurrentMessageNumber: TMsgNumber;
begin
  Result := DoGetCurrentMessage;
  if Result = TMsgNumber_FAIL then
  begin
    FmxErrorCheck;
  end;
end;

function TFmxCustomMessageArea.GetFirstMessageNumber: TMsgNumber;
begin
  Result := DoGetFirstMessage;
  if Result = TMsgNumber_FAIL then
  begin
    FmxErrorCheck;
  end;
end;

function TFmxCustomMessageArea.GetHighestReadMessageId: UMSG;
var
  lri: LRINFO;
begin
  if not DoGetLRInfo(lri) then FmxErrorCheck;
  Result := lri.umsgHigh;
end;

function TFmxCustomMessageArea.GetHighMessageNumber: TMsgNumber;
begin
  Result := DoGetHighMessage;
  if Result = TMsgNumber_FAIL then
  begin
    FmxErrorCheck;
  end;
end;

function TFmxCustomMessageArea.GetLastReadMessageId: UMSG;
var
  lri: LRINFO;
begin
  if not DoGetLRInfo(lri) then FmxErrorCheck;
  Result := lri.umsgLast;
end;

function TFmxCustomMessageArea.GetMessage(msgn: TMsgNumber;
  Mode: TFmxMessageOpenMode;
  Flags: TFmxMessageOpenFlags): TFmxCustomAreaMessage;
begin
  Result :=
    DoGetMessage(
      msgn,
      FmxMapMessageOpenModeToFmaMessageOpenMode(Mode),
      FmxMapMessageOpenFlagsToFmaMessageOpenFlags(Flags)
      );
end;

function TFmxCustomMessageArea.GetMessageCount: LONG;
begin
  Result := DoGetMessageCount;
  if Result = 0 then // can be error
  begin
    FmxErrorCheck;
  end;
end;

{$IFDEF FMA_D5}
function TFmxCustomMessageArea.GetMessageMap: TMsgIdArray;
var
  aumsg: PUMSG;
  Count: ULONG;
  I: Integer;
begin
  Result := nil;
  if not DoCreateMsgMap(aumsg, Count) then FmxErrorCheck;
  if Count > 0 then
  begin
    SetLength(Result, Count);
    for I := 0 to Pred(Count) do
    begin
      Result[I] := FmaDef.AUMSG(aumsg)[I];
    end;
    FmaMemFree(aumsg);
  end;
end;
{$ENDIF}
procedure TFmxCustomMessageArea.Last;
begin
  SetCurrentMessage(GetHighMessageNumber);
end;

procedure TFmxCustomMessageArea.Lock;
begin
  if not DoLock(True) then FmxErrorCheck;
end;

function TFmxCustomMessageArea.MsgIdToMsgNumber(Value: TMsgId; Mode: TFmxMsgIdConvertionMode): TMsgNumber;
begin
  Result := DoMsgIdToMsgNumber(Value, FmxMapMsgIdConvertionModeToFmaMsgIdConvertionMode(Mode));
  if Result = TMsgNumber_FAIL then
  begin
    FmxErrorCheck;
  end;
end;

function TFmxCustomMessageArea.MsgNumberToMsgId(Value: TMsgNumber): TMsgId;
begin
  Result := DoMsgNumberToMsgId(Value);
  if Result = NoMsgId then
  begin
    FmxErrorCheck;
  end;
end;

procedure TFmxCustomMessageArea.Next;
begin
  SetCurrentMessage(Succ(MessageNumber));
end;

procedure TFmxCustomMessageArea.Open;
var
  TempAddr: TFmxNetAddr;
begin
  Close;
  TempAddr := NetAddr;
  DoOpen(
    PAnsiChar(Path),
    FmxMapAreaOpenModeToFmaAreaOpenMode(OpenMode),
    FmxMapAreaOpenFlagsToFmaOpenAreaFlags(OpenFlags),
    TempAddr
    );
  if FHandle = NoAreaHandle then
  begin
    FmxErrorCheck;
  end;
  if MessageCount > 0 then
  begin
    First;
  end;
end;

procedure TFmxCustomMessageArea.Prev;
begin
  SetCurrentMessage(Pred(MessageNumber));
end;

procedure TFmxCustomMessageArea.SetActive(const Value: Boolean);
begin
  if Value <> Active then
  begin
    if Value then
    begin
      Open;
    end
    else
    begin
      Close;
    end;
  end;
end;

procedure TFmxCustomMessageArea.SetCurrentMessage(msgn: TMsgNumber);
begin
  if not DoSetCurrentMessage(msgn) then
  begin
    FmxErrorCheck;
  end;
end;

procedure TFmxCustomMessageArea.SetFlags(const Value: TFmxAreaOpenFlags);
begin
  if FFlags <> Value then
  begin
    FFlags := Value;
    Active := False;
  end;
end;

procedure TFmxCustomMessageArea.SetHandle(Value: TAreaHandle);
begin
  FHandle := Value;
end;

procedure TFmxCustomMessageArea.SetHighestReadMessageId(const Value: UMSG);
var
  lri: LRINFO;
begin
  if not DoGetLRInfo(lri) then
  begin
    lri.umsgLast := Value;
  end;
  lri.umsgHigh := Value;
  if not DoSetLRInfo(lri) then FmxErrorCheck;
end;

procedure TFmxCustomMessageArea.SetLastReadMessageId(const Value: UMSG);
var
  lri: LRINFO;
begin
  if not DoGetLRInfo(lri) then
  begin
    lri.umsgHigh := Value;
  end;
  lri.umsgLast := Value;
  if not DoSetLRInfo(lri) then FmxErrorCheck;
end;

procedure TFmxCustomMessageArea.SetNetAddr(const Value: TFmxNetAddr);
begin
  if not SameNetAddr(FNetAddr, Value) then
  begin
    FNetAddr := Value;
    Active := False;
  end;
end;

procedure TFmxCustomMessageArea.SetOpenMode(const Value: TFmxAreaOpenMode);
begin
  if FOpenMode <> Value then
  begin
    FOpenMode := Value;
    Active := False;
  end;
end;

procedure TFmxCustomMessageArea.SetPath(const Value: String);
begin
  if FPath <> Value then
  begin
    FPath := Value;
    Active := False;
  end;
end;

procedure TFmxCustomMessageArea.Unlock;
begin
  if not DoLock(False) then FmxErrorCheck;
end;

function TFmxCustomMessageArea.Validate: ULONG;
begin
  ValidateMessageNumber(MessageNumber, Result);
end;

function TFmxCustomMessageArea.ValidateMessageNumber(MsgN: TMsgNumber;
  var MsgIndex: ULONG): Boolean;
begin
  Result := DoValidateMessageNumber(MsgN, MsgIndex);
  if not Result then FmxErrorCheckExcept([FMAERROR_NOT_FOUND]);
end;

{ TFmxCustomAreaMessage }

procedure TFmxCustomAreaMessage.Close;
begin
  if Handle <> NoMsgHandle then
  begin
    if not DoClose then FmxErrorCheck;
    FHandle := NoMsgHandle;
  end;
end;

constructor TFmxCustomAreaMessage.CreateOn(MsgHandle: TMsgHandle);
begin
  if MsgHandle = NoMsgHandle then
  begin
    FmxErrorCheck;
  end;
  Inherited Create;
  FHandle := MsgHandle;
end;

destructor TFmxCustomAreaMessage.Destroy;
begin
  Close;
  inherited Destroy;
end;

function TFmxCustomAreaMessage.GetControlInfoLen: ULONG;
begin
  if not DoGetControlInfoLen(Result) then
  begin
    FmxErrorCheck;
  end;
end;

function TFmxCustomAreaMessage.GetTextLen: ULONG;
begin
  if not DoGetTextLen(Result) then
  begin
    FmxErrorCheck;
  end;
end;



{
  NB (Andrew D. Plyako)
  -   --  
  - temporary method -- clear kludges
}
procedure RemoveKludges(var Str: String);
var
  kludgePos, i, leng: integer;
  c: char;
begin
  kludgePos := Pos(#01,str);
  while kludgePos<>0 do
    begin
     leng := Length(str); c := Str[kludgePos]; i:= kludgePos+1;
     while (i<=leng) or ((str[i]=#10) and (c=#13)) do
       begin c:= str[i]; inc(i) end;
     if i>leng then dec(i);
     Delete(str,kludgePos,i-kludgePos+1);
     kludgePos := Pos(#01,str);
    end;
end;
// NB (Andrew D. Plyako)

{
  NB (Andrew D. Plyako)
  -       PropAccessor
  - New method: read message text
// Note:     ...   
// Note: Kludges do not read
}
procedure TFmxCustomAreaMessage.Read(lines: TStrings);
var
  pchCtrl, pchText: PAnsiChar;
  cchText, cchCtrl: ULONG;
  linesText: String;
begin
  cchText:=FRM_GETALLTEXT; cchCtrl:=FRM_GETCTRLPTR;
  pchText:=nil; pchCtrl:=nil;
  if not DoRead(nil,pchCtrl,Addr(cchCtrl),pchText,Addr(cchText),nil) then
    FmxErrorCheck;
  try
    SetLength(linesText, cchText);
    CopyMemory(PChar(Addr(linesText[1])), pchText, cchText);
  finally
    FmaMemFree(pchText);
  end;
  RemoveKludges(linesText);
  lines.Text := Trim(linesText);
end;
// NB (Andrew D. Plyako)


function TFmxCustomAreaMessage.Load(fl: ULONG): PMsgObject;
begin
  if not DoLoad(Result, fl) then
  begin
    FmxErrorCheck;
  end;
end;

procedure TFmxCustomAreaMessage.Save(pfmm: PMsgObject; fl: ULONG);
begin
  if not DoSave(pfmm, fl, ULONG(-1)) then
  begin
    FmxErrorCheck;
  end;
end;

procedure TFmxCustomAreaMessage.SaveLen(pfmm: PMsgObject; fl,
  TextTotalLen: ULONG);
begin
  if not DoSave(pfmm, fl, TextTotalLen) then
  begin
    FmxErrorCheck;
  end;
end;

procedure TFmxCustomAreaMessage.SetHandle(Value: TMsgHandle);
begin
  FHandle := Value;
end;

{ TFmxMessageObject }

function TFmxMessageObject.AddASZProperty(Id: ULONG;
  Value: PAnsiChar): PFMP;
begin
  Result := AddProperty(Id, Value, ULONG(-1));
end;

function TFmxMessageObject.AddBINProperty(Id: ULONG;
  Values: array of BYTE): PFMP;
begin
  if High(Values) >= 0 then
  begin
    Result := AddProperty(Id, Addr(Values[0]), Succ(High(Values)) * sizeof(BYTE));
  end
  else
  begin
    Result := AddProperty(Id, nil, 0);
  end;
end;

function TFmxMessageObject.AddDTMProperty(Id: ULONG;
  Value: TDateTime): PFMP;
var
  D: DTTM;
begin
  D := DateTimeToDTTM(Value);
  Result := AddProperty(Id, Addr(D), sizeof(D));
end;

function TFmxMessageObject.AddProperty(Id: ULONG; Data: Pointer;
  Size: ULONG): PFMP;
begin
  Result := FmaAddFmp(DataPointer, Id, Data, Size);
  if not Assigned(Result) then
  begin
    FmxErrorCheck;
  end;
end;

function TFmxMessageObject.AddPSZProperty(Id: ULONG;
  Values: array of PAnsiChar): PFMP;
begin
  if High(Values) >= 0 then
  begin
    Result := AddProperty(Id, Addr(Values[0]), Succ(High(Values)) * sizeof(PAnsiChar));
  end
  else
  begin
    Result := AddProperty(Id, nil, 0);
  end;
end;

{$IFDEF FMA_D5}
function TFmxMessageObject.AddStringArrayProperty(Id: ULONG;
  Values: array of AnsiString): PFMP;
var
  AV: array of PAnsiChar;
  I: Integer;
  L: Integer;
begin
  AV := nil;
  if High(Values) >= 0 then
  begin
    SetLength(AV, Succ(High(Values)));
    for I := 0 to High(Values) do
    begin
      if Values[I] <> '' then
      begin
        L := Length(Values[I]) + 1;
        AV[I] := FmaMemAlloc(L, False);
        CopyMemory(AV[I], PAnsiChar(Values[I]), L);
      end
      else
      begin
        AV[I] := nil;
      end;
    end;
  end;
  Result := AddPSZProperty(Id, AV);
end;
{$ENDIF}

function TFmxMessageObject.AddStringProperty(Id: ULONG;
  Value: AnsiString): PFMP;
begin
  if Value <> '' then
    Result := AddProperty(Id, PAnsiChar(Value), Length(Value) + 1)
  else
    Result := AddProperty(Id, nil, 0);
end;

{$IFDEF FMA_D5}
{
  NB (Andrew D. Plyako)
  - ,  nil'     #13#10
  - fixed. Empty strings should be #13#10 (not nil)
}
function TFmxMessageObject.AddStringsProperty(Id: ULONG;
  List: TStrings): PFMP;
var
  AV: array of PAnsiChar;
  I: Integer;
  L: Integer;
begin
  AV := nil;
  if Assigned(List) and (List.Count > 0) then
  begin
    SetLength(AV, List.Count);
    for I := 0 to Pred(List.Count) do
    begin
      L := Length(List[I]);
      AV[I] := FmaMemAlloc(L+3, False);
      if L <> 0 then
        CopyMemory(AV[I], PAnsiChar(List[I]), L);
      AV[i][L]  :=#13;
      AV[i][L+1]:=#10;
      AV[i][L+2]:=#0;
    end;
  end;
  Result := AddPSZProperty(Id, AV);
end;
// NB (Andrew D. Plyako)
{$ENDIF}

function TFmxMessageObject.AddU32Property(Id: ULONG;
  Values: array of LONG): PFMP;
begin
  if High(Values) >= 0 then
  begin
    Result := AddProperty(Id, Addr(Values[0]), Succ(High(Values)) * sizeof(LONG));
  end
  else
  begin
    Result := AddProperty(Id, nil, 0);
  end;
end;

constructor TFmxMessageObject.CreateClone(AnObject: TFmxMessageObject);
var
  AData: PMsgObject;
begin
  AData := FmaDupeMsg(AnObject.DataPointer);
  if not Assigned(AData) then
  begin
    FmxErrorCheck;
  end;
  Inherited Create;
  FData := AData;
  FAccessor := nil;
end;

constructor TFmxMessageObject.CreateNew(ExtraSize: UINT);
var
  AData: PMsgObject;
begin
  AData := FmaCreateMsg(ExtraSize);
  if not Assigned(AData) then
  begin
    FmxErrorCheck;
  end;
  Inherited Create;
  FData := AData;
  FAccessor := nil;
end;

constructor TFmxMessageObject.CreateOn(AData: PMsgObject);
begin
  if not Assigned(AData) then
  begin
    FmxErrorCheck;
  end;
  Inherited Create;
  FData := AData;
  FAccessor := nil;
end;

destructor TFmxMessageObject.Destroy;
begin
  FAccessor.Free;
  FmaDestroyMsg(DataPointer);
  inherited Destroy;
end;

procedure TFmxMessageObject.Dump;
begin
  FmaDumpMsg(DataPointer);
end;

function TFmxMessageObject.GetAttributes: TFmxMsgAttributes;
begin
  Result := FmxMapMsgAttrToMsgAttributes(DataPointer^.attr);
end;

{
  NB (Andrew D. Plyako)
  -  Attributes    
  - property "Attributes" is writeable now
}
procedure TFmxMessageObject.SetAttributes(const Value: TFmxMsgAttributes);
begin
  DataPointer^.attr := FmxMapMsgAttributesToMsgAttr(Value);
end;
// NB (Andrew D. Plyako)


function TFmxMessageObject.GetFirstProperty: PFMP;
begin
  Result := DataPointer^.pfmp;
end;

function TFmxMessageObject.GetNativeType: TFmxMsgType;
begin
  Result := FmxMapFmaMsgTypeToMsgType(DataPointer^._type);
end;

function TFmxMessageObject.GetPropAccessor: TFmxMsgPropAccessor;
begin
  if not Assigned(FAccessor) then
  begin
    FAccessor := TFmxMsgPropAccessor.Create(Self);
  end;
  Result := FAccessor;
end;

function TFmxMessageObject.GetProperty(Id: ULONG; Index: UINT): PFMP;
begin
  Result := FmaGetFmp(DataPointer, Id, Index);
  if not Assigned(Result) then
  begin
    FmxErrorCheck;
  end;
end;

procedure TFmxMessageObject.RemoveProperties(Id: ULONG);
var
  Did: Boolean;
begin
{$IFDEF FMA_D5}
  FreeAndNil(FAccessor);
{$ELSE}
  FAccessor.Free;
  FAccessor:=nil;
{$ENDIF}
  while True do
  begin
    Did := FmaDelFmp(DataPointer, Id, 0);
    if not Did then
    begin
      FmxErrorCheckExcept([FMAERROR_NOT_FOUND]);
      break;
    end;
  end;
end;

function TFmxMessageObject.RemoveProperty(Id: ULONG; Index: UINT): Boolean;
begin
{$IFDEF FMA_D5}
  FreeAndNil(FAccessor);
{$ELSE}
  FAccessor.Free;
  FAccessor:=nil;
{$ENDIF}
  Result := FmaDelFmp(DataPointer, Id, Index);
  if not Result then FmxErrorCheckExcept([FMAERROR_NOT_FOUND]);
end;

{ TFmxMsgPropAccessor }

constructor TFmxMsgPropAccessor.Create(AMsgObj: TFmxMessageObject);
begin
  if not Assigned(AMsgObj) or not Assigned(AMsgObj.DataPointer) then
  begin
    FmxErrorRaise(FMAERROR_INVALID_PARAM);
  end;
  Inherited Create;
  FMsgObj := AMsgObj.DataPointer;
  First;
end;

constructor TFmxMsgPropAccessor.CreateOnPFMP(AMsgObj: PMsgObject);
begin
  if not Assigned(AMsgObj) then
  begin
    FmxErrorRaise(FMAERROR_INVALID_PARAM);
  end;
  Inherited Create;
  FMsgObj := AMsgObj;
  First;
end;

procedure TFmxMsgPropAccessor.Delete;
var
  Curr: PFMP;
  Id: ULONG;
  Index: UINT;
begin
  Curr := FMsgObj^.pfmp;
  Index := UINT(-1);
  Id := FData^.id;
  while Assigned(Curr) do
  begin
    if Curr^.id = Id then
    begin
      Inc(Index);
    end;
    if Curr = FData then
    begin
      Next; // set FData to next property
      if not FmaDelFmp(FMsgObj, Id, Index) then
      begin
        FmxErrorCheck;
      end;
      Exit;
    end;
    Curr := Curr^.pfmpNext;
  end;
  FmxErrorRaise(FMAERROR_INVALID_DATA);
end;

procedure TFmxMsgPropAccessor.Dump;
begin
  FmaDumpFmp(FData);
end;

function TFmxMsgPropAccessor.Eol: Boolean;
begin
  Result := not Assigned(FData);
end;

procedure TFmxMsgPropAccessor.First;
begin
  FData := FMsgObj^.pfmp;
end;

function TFmxMsgPropAccessor.GetASZ: PAnsiChar;
begin
  Result := PAnsiChar(PAnsiChar(FData) + sizeof(FMP));
end;

function TFmxMsgPropAccessor.GetByte(Index: Cardinal): BYTE;
begin
  Result := PBYTE(PAnsiChar(FData) + sizeof(FMP) + (Index * sizeof(BYTE)))^;
end;

function TFmxMsgPropAccessor.GetByteCount: Integer;
begin
  Result := FData.cb;
end;

function TFmxMsgPropAccessor.GetDateTime: TDateTime;
var
  D: DTTM;
begin
  D := PDTTM(PAnsiChar(FData) + sizeof(FMP))^;
  Result := DTTMtoDateTime(D);
end;

function TFmxMsgPropAccessor.GetId: ULONG;
begin
  Result := FData^.id;
end;

function TFmxMsgPropAccessor.GetPosition: Byte;
begin
  Result := (FData^.id and FMP_POSMASK) shr 16;
end;

function TFmxMsgPropAccessor.GetPropertyData: Pointer;
begin
  Result := Pointer(PAnsiChar(FData) + sizeof(FMP));
end;

function TFmxMsgPropAccessor.GetPSZ(Index: Cardinal): PAnsiChar;
type
  PPChar=^PChar;
begin
{$IFDEF FMA_D5}
  Result := APAnsiChar(PAnsiChar(FData) + sizeof(FMP))[Index];
{$ELSE}
  Result := PPChar(DWORD(FData) + 
             (sizeof(FMP) + Index*sizeof(PPChar)))^;
{$ENDIF}
end;

function TFmxMsgPropAccessor.GetPSZCount: Integer;
begin
  Result := FData.cb div sizeof(PAnsiChar);
end;

function TFmxMsgPropAccessor.GetString: AnsiString;
var
  List: TStrings;
begin
  List := TStringList.Create;
  try
    GetStrings(List);
    Result := List.Text;
  finally
    List.Free;
  end;
end;

procedure TFmxMsgPropAccessor.GetStrings(List: TStrings);
var
  I: Integer;
begin
  List.BeginUpdate;
  try
    List.Clear;
    if IsASZ then
    begin
      List.Add(GetASZ);
    end
    else
    if IsPSZs then
    begin
      for I := 0 to Pred(PSZCount) do
      begin
        List.Add(GetPSZ(I));
      end;
    end
    else
    begin
      FmxErrorRaise(FMAERROR_INVALID_PARAM);
    end;
  finally
    List.EndUpdate;
  end;
end;

function TFmxMsgPropAccessor.GetTag: UINT;
begin
  Result := UINT(FData^.id and FMP_TAGMASK);
end;

function TFmxMsgPropAccessor.GetU32(Index: Cardinal): LONG;
begin
{$IFDEF FMA_D5}
  Result := ALONG(PAnsiChar(FData) + sizeof(FMP))[Index];
{$ELSE}
  Result := PLONG(DWORD(FData) + 
             (sizeof(FMP) + Index*sizeof(PLONG)))^;
{$ENDIF}
end;
function TFmxMsgPropAccessor.GetU32Count: Integer;
begin
  Result := FData.cb div sizeof(LONG);
end;

function TFmxMsgPropAccessor.IsASZ: Boolean;
begin
  Result := ((FData^.id and FMP_ASZ) = FMP_ASZ);
end;

function TFmxMsgPropAccessor.IsBytes: Boolean;
begin
  Result := ((FData^.id and FMP_BIN) = FMP_BIN);
end;

function TFmxMsgPropAccessor.IsDateTime: Boolean;
begin
  Result := ((FData^.id and FMP_DTM) = FMP_DTM);
end;

function TFmxMsgPropAccessor.IsEndOfMessage: Boolean;
begin
  Result := ((FData^.id and FMP_EOM) = FMP_EOM);
end;

function TFmxMsgPropAccessor.IsKludge: Boolean;
begin
  Result := ((FData^.id and FMP_KLU) = FMP_KLU);
end;

function TFmxMsgPropAccessor.IsPSZs: Boolean;
begin
  Result := ((FData^.id and FMP_PSZ) = FMP_PSZ);
end;

function TFmxMsgPropAccessor.IsSingleInstance: Boolean;
begin
  Result := ((FData^.id and FMP_ONE) = FMP_ONE);
end;

function TFmxMsgPropAccessor.IsU32s: Boolean;
begin
  Result := ((FData^.id and FMP_U32) = FMP_U32);
end;

procedure TFmxMsgPropAccessor.Last;
var
  Prev: PFMP;
  Curr: PFMP;
begin
  Prev := nil;
  Curr := FMsgObj^.pfmp;
  while Assigned(Curr) do
  begin
    Prev := Curr;
    Curr := Curr^.pfmpNext;
  end;
  FData := Prev;
end;

procedure TFmxMsgPropAccessor.Next;
begin
  if not Eol then
  begin
    FData := FData^.pfmpNext;
  end;
end;

procedure TFmxMsgPropAccessor.Prev;
begin
  if not Eol then
  begin
    FData := FData^.pfmpPrev;
  end;
end;

procedure TFmxMsgPropAccessor.SetByte(Index: Cardinal; Value: BYTE);
begin
  PBYTE(PAnsiChar(FData) + sizeof(FMP) + (Index * sizeof(BYTE)))^ := Value;
end;

procedure TFmxMsgPropAccessor.SetDateTime(Value: TDateTime);
var
  D: DTTM;
begin
  D := DateTimeToDTTM(Value);
  PDTTM(PAnsiChar(FData) + sizeof(FMP))^ := D;
end;

procedure TFmxMsgPropAccessor.SetId(const Value: ULONG);
begin
  FData^.id := Value;
end;

procedure TFmxMsgPropAccessor.SetPosition(const Value: Byte);
begin
  FData^.id := (FData^.id and not FMP_POSMASK) or (ULONG(Value) shl 16);
end;

procedure TFmxMsgPropAccessor.SetPSZ(Index: Cardinal; Value: AnsiString);
var
  P: PAnsiChar;
  L: Integer;
begin
  if Value <> '' then
  begin
    L := Length(Value) + 1;
    P := FmaMemAlloc(L, False);
    CopyMemory(P, PAnsiChar(Value), L);
  end
  else
  begin
    P := nil;
  end;
  PPAnsiChar(PAnsiChar(FData) + sizeof(FMP) + (Index * sizeof(PAnsiChar)))^ := P;
end;

procedure TFmxMsgPropAccessor.SetTag(const Value: UINT);
begin
  FData^.id := (FData^.id and not FMP_TAGMASK) or Value;
end;

procedure TFmxMsgPropAccessor.SetU32(Index: Cardinal; Value: LONG);
begin
  PLONG(PAnsiChar(FData) + sizeof(FMP) + (Index * sizeof(LONG)))^ := Value;
end;


{
  NB (Andrew D. Plyako)
  -  FData      ID
  - Scan through FData for property with ID identificator
}
function TFmxMsgPropAccessor.FindProperty(const PropID: ULONG): Boolean;
begin
  First;
  while Assigned(FData) and (FData^.ID<>PropID) do Next;
  Result := Assigned(FData) and (FData^.ID=PropID)
end;
// NB (Andrew D. Plyako)


{
  NB (Andrew D. Plyako)
  -  FData   Kludge;    FMA
  - Scan through FData for the Kludge; this method patch FMA bug
}
function TFmxMsgPropAccessor.FindKludge(Kludge: string): Boolean;
var
  CurKludge: string;
  klLeng: Integer;
begin
  Result := false;
  Kludge := UpperCase(Kludge); klLeng := Length(Kludge);
  First;
  repeat
    while Assigned(FData) and not IsKludge do Next;
    if Assigned(FData) then
      begin
       CurKludge := UpperCase(GetString);
       Result := (Length(CurKludge) >= klLeng) and
                 (Copy(CurKludge,1,klLeng) = Kludge);
       if not Result then Next;
      end;
  until Result or not Assigned(FData);
end;
// NB (Andrew D. Plyako)


{
  NB (Andrew D. Plyako)
  -  IgnoreErrors
  - Initialize IgnoreErrors
}
initialization
  IgnoreErrors := [];
// NB (Andrew D. Plyako)
  
end.
